/*  DeltaSMT: Delta-debugger for Satisfiablity Modulo Theories (SMT) benchmarks.
 *  Copyright (C) 2009  Robert Daniel Brummayer
 *
 *  This file is part of DeltaSMT.
 *
 *  DeltaSMT is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DeltaSMT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;

public class Main
{

  private static String g_tmpName = "DeltaSMTtmp.smt";

  private static String g_version = "0.1";

  private static final String header =
"********************************************************************************\n" +
"*              DeltaSMT " + g_version + "                                                    *\n"  +
"*              Delta-Debugger for SMT-LIB 1.2 Benchmarks                       *\n" +
"*              written by Robert Daniel Brummayer, 2009                        *\n" + 
"********************************************************************************\n" +
"\n";

  private static void printHelp(){
    System.out.print (header);
    System.out.print("deltasmt [<option>...] <input> <output> ");
    System.out.println("<exec> [<exec-arg>...]");
    System.out.println("");
    System.out.println("where <option> is one of the following:");
    System.out.println("");
    System.out.println(" -h               print this help screen and exit");
    System.out.println(" -V               print version and exit");
    System.out.println(" -v               increase verbosity (max 3)");
    System.out.println(" -t <seconds>     use <seconds> timeout for each call");
    System.out.println("");
    System.out.println(" -no-norm         do not normalize during structural" + 
                                          " hashing");
    System.out.println(" -no-colors       disable colors in verbose messages");
    System.out.println("");
    System.out.println("supported logics:");
    System.out.println("  QF_A, QF_AX, QF_BV, QF_AUFBV, QF_AUFLIA, QF_IDL, " +
                       "QF_LIA, QF_NIA,");
    System.out.println("  QF_RDL, QF_LRA, QF_NRA, QF_UF, QF_UFIDL, QF_UFLIA, "
                       + "QF_UFNIA,"); 
    System.out.println("  QF_UFRDL, QF_UFLRA, QF_UFNRA");
    System.out.println("");
  }

  public static void main (String args[]) {
    DeltaSMT delta;
    int i;
    int verbosity = 0;
    boolean col = true;
    boolean normSharing = true;
    long timeout= 604800000; /* one week in milliseconds */
    String arg;
    FileReader in = null;
    PrintStream out = null;
    String []cmdArray = null;
    String inName = null;
    String outName = null;
    
    if (args.length < 3 && !(args.length == 1 && 
        (args[0].equals("-V") || args[0].equals ("-h")))){
      printHelp();
      System.exit(1);
    }

    for (i = 0; i < args.length; i++){
      arg = args[i];
      if (arg.charAt(0) == '-'){
        if (arg.equals("-V")){
          System.out.println(g_version);
          System.exit (0);
        } else if (arg.equals("-h")){
          printHelp ();
          System.exit(0);
        } else if (arg.equals("-v")){
          verbosity++;
        } else if (arg.equals("-no-colors")){
          col = false;
        } else if (arg.equals("-no-norm")){
          normSharing = false;
        } else if (arg.equals("-t")){
          if (i == args.length - 1){
            System.out.println ("argument for timeout missing");
            System.exit (1);
          }
          i++;
          try {
            timeout = Long.parseLong(args[i]);
          } catch (NumberFormatException nfe) {
            System.out.println ("invalid argument for timeout");
            System.exit (1);
          }
          if (timeout <= 0l){
            System.out.println ("timeout must be greater than zero");
            System.exit (1);
          }
          /* check if conversion in milliseconds overflows */
          timeout = timeout * 1000l;
          if (timeout <= 0l){
            System.out.println ("timeout is too large");
            System.exit (1);
          }
        } else {
          System.err.println("invalid option: \"" + arg + "\"");
          System.exit(1);
        }
      } else {
        /* check if input and output files can be read resp. written */
        if (inName == null){
          try {
            in = new FileReader(arg);
          } catch (FileNotFoundException fnfe){
            System.err.println("cannot read input: " + fnfe.getMessage());
            System.exit(1);
          }
          try {
            in.close();
          } catch (IOException ioe) { /* ignore */ }
          inName = arg;
        } else if (outName == null){
          try {
            out = new PrintStream(arg);
          } catch (FileNotFoundException fnfe){
            System.err.println("cannot write output: " + fnfe.getMessage());
            System.exit(1);
          }
          out.close();
          outName = arg;
        } else {
          break;
        }
      }
    }

    cmdArray = new String[args.length - i];
    for (int j = 0; j < cmdArray.length; j++)
      cmdArray[j] = args[j + i];

    if (inName == null || outName == null || cmdArray  == null){
      printHelp();
      System.exit(1);
    }

    delta = new DeltaSMT(inName, outName, g_tmpName, cmdArray, verbosity, col,
                         timeout, normSharing);
    try {
      delta.deltaDebug();
    } catch (StackOverflowError sofe){
      System.err.println("stack overflow while parsing");
      System.err.println("increase the stack size of the virtual machine");
      System.exit(1);
    } catch (Exception e) {
      System.out.print("exception occurred");
      if (e.getMessage() != null)
        System.out.println(": " + e.getMessage());
      else
        System.out.println("");
      /* e.printStackTrace(); */
      System.exit(1);
    }

    System.exit(0);
  }

}
