/*  DeltaSMT: Delta-debugger for Satisfiablity Modulo Theories (SMT) benchmarks.
 *  Copyright (C) 2009  Robert Daniel Brummayer
 *
 *  This file is part of DeltaSMT.
 *
 *  DeltaSMT is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DeltaSMT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.*;
import java.io.*;

public class SMTBenchmark
{

  protected SMTNodeManager mgr;

  protected String name;
  
  protected SMTLogic logic;

  protected SMTBenchmarkStatus status;

  protected List<SMTNode> assumptions;

  protected SMTNode formula;

  protected HashMap<String,UType> extraSorts;


  public SMTBenchmark (SMTNodeManager mgr){
    assert (mgr != null);
    this.mgr = mgr;
    this.name = null;
    this.logic = null;
    this.status = null;
    this.assumptions = null;
    this.formula = null;
    this.extraSorts = null;
  } 

  public String getName(){
    return this.name;
  }

  public void setName(String name){
    assert (name != null);
    this.name = name;
  }

  public SMTLogic getLogic(){
    return this.logic;
  }

  public void setLogic(SMTLogic logic){
    assert (logic != null);
    this.logic = logic;
  }

  public SMTBenchmarkStatus getStatus(){
    return this.status;
  }

  public void setStatus(SMTBenchmarkStatus status){
    assert (status != null);
    this.status = status;
  }

  public SMTNode getRootFormula(){
    return this.formula;
  }

  public void setRootFormula(SMTNode formula){
    assert (formula != null);
    this.formula = formula;
  }

  public HashMap<String,UType> getExtraSorts () {
    return this.extraSorts;
  }

  public void setExtraSorts (HashMap<String,UType> extraSorts){
    this.extraSorts = extraSorts;
  }

  public List<SMTNode> getAssumptions () {
    return this.assumptions;
  }

  public void setAssumptions (List<SMTNode> assumptions) {
    assert (assumptions != null);
    this.assumptions = assumptions;
  }

  public void release(){
    if (this.assumptions != null)
      for (SMTNode assumption : this.assumptions)
        this.mgr.release (assumption);

    if (this.formula != null)
      this.mgr.release (this.formula);
  }

  protected void dumpPreamble (PrintStream out, HashMap<Long,SMTNode> subst) {
    Extrafun fun;
    Stack<SMTNode> stack;
    SMTNode cur, root, temp;
    SMTNode [] children;
    HashSet<String> stringHash = new HashSet<String>();
    StringBuilder builder;
    String decl;

    assert (out != null);
    assert (subst != null);

    out.println ("(benchmark " + this.name);
    out.println(":logic " + this.logic.toString());

    assert (this.extraSorts != null);
    for (String s : this.extraSorts.keySet())
      out.println (":extrasorts (" + s + ")");
    
    stack = new Stack<SMTNode>();
    cur = subst.get (this.formula.getId());
    if (cur == null)
      root = cur = this.formula;
    else
      root = cur;
    stack.push(cur);
    while (!stack.empty()) {
      cur = stack.pop();
      assert (cur.getRefs() > 0);
      if ((temp = subst.get(cur.getId())) != null)
        cur = temp;
      assert (cur.getMark() == 0 || cur.getMark() == 1);
      if (cur.getMark() == 1)
        continue;
      assert (cur.getMark() == 0);
      cur.setMark((byte) 1);
      children = cur.getChildren();
      if (cur.getKind() == SMTNodeKind.EXTRAFUN) {
        builder = new StringBuilder();
        builder.append (":extrafuns ((");
        builder.append (((Extrafun) cur).toDeclString());
        builder.append ("))");
        decl = builder.toString();
        if (!stringHash.contains(decl)) {
          out.println (decl);
          stringHash.add(decl);
        }
      } else if (cur.getKind() == SMTNodeKind.EXTRAPRED){
        builder = new StringBuilder();
        builder.append (":extrapreds ((");
        builder.append (((Extrapred) cur).toDeclString());
        builder.append ("))");
        decl = builder.toString();
        if (!stringHash.contains(decl)) {
          out.println (decl);
          stringHash.add(decl);
        }
      }
      if (children != null)
        for (SMTNode child : children)
          stack.push (child);

    }

    SMTNodeManager.setMark((byte) 0, root);
    /* clear mark flag of substitutions */
    for (SMTNode n : subst.values())
      SMTNodeManager.setMark((byte) 0, n);


    if (this.status != null)
      out.println(":status " + this.status.getString());
    out.println(":formula");

  }

  /* substitutions are not allowed to be cyclic,
   * this code does not dump assumptions, they are merged
   * into the main formula by the delta-debugger */
  public void dumpSubstitute (PrintStream out, HashMap<Long, SMTNode> subst){
    Stack<SMTNode> stack;
    HashMap<Long,String> table;
    SMTNode cur, temp, root;
    SMTNode []children;
    int id;
    String str;

    assert (out != null);
    assert (this.formula != null);
    assert (this.formula.getType() == SMTTypeManager.booleanType);
    assert (subst != null);

    id = 1;
    stack = new Stack<SMTNode>();
    table = new HashMap<Long,String>();

    cur = subst.get (this.formula.getId());
    if (cur == null)
      root = cur = this.formula;
    else
      root = cur;

    this.dumpPreamble(out, subst);

    if (cur == SMTNodeManager.TRUE) {
      out.println ("true");
      out.println (")");
      return;
    }

    if (cur == SMTNodeManager.FALSE) {
      out.println ("false");
      out.println (")");
      return;
    }

    stack.push(cur);
    while (!stack.empty()) {
      cur = stack.pop();
      assert (cur.getRefs() > 0);

      if ((temp = subst.get(cur.getId())) != null)
        cur = temp;
      assert (cur.getRefs() > 0);
      assert (cur.getMark() >= 0 && cur.getMark() <= 2);

      if (cur.getMark() == 2)
        continue;

      if (cur.getMark() == 0){
        switch (cur.getKind()){
          case TRUE:
            str = "$n" + id++;
            out.println("(flet (" + str + " true)");
            cur.setMark((byte) 2);
            table.put (cur.getId(), str);
            break;
          case FALSE:
            str = "$n" + id++;
            out.println("(flet (" + str + " false)");
            cur.setMark((byte) 2);
            table.put (cur.getId(), str);
            break;
          case CONST:
            str = "?n" + id++;
            out.println("(let (" + str + " " + cur.toString() + ")");
            cur.setMark((byte) 2);
            table.put (cur.getId(), str);
            break;
          case EXTRAFUN:
            if (((Extrafun) cur).isVar()) {
              cur.setMark((byte) 2);
              table.put (cur.getId(), ((Extrafun) cur).getName());
            } else {
              cur.setMark((byte) 1);
              children = cur.getChildren();
              assert (children != null);
              stack.push (cur);
              for (int i = children.length - 1; i >= 0; i--)
                stack.push (children[i]);
            }
            break;
          case EXTRAPRED:
            if (((Extrapred) cur).isVar()) {
              cur.setMark((byte) 2);
              table.put (cur.getId(), ((Extrapred) cur).getName());
            } else {
              cur.setMark((byte) 1);
              children = cur.getChildren();
              assert (children != null);
              stack.push (cur);
              for (int i = children.length - 1; i >= 0; i--)
                stack.push (children[i]);
            }
            break;
          default:
            cur.setMark((byte) 1);
            children = cur.getChildren();
            assert (children != null);
            stack.push (cur);
            for (int i = children.length - 1; i >= 0; i--)
              stack.push (children[i]);
            break;
        }
      } else {
        assert (cur.getMark() == 1);
        cur.setMark((byte) 2);
        if (cur.getType() == SMTTypeManager.booleanType){
          str = "$n" + id++;
          out.print("(flet (" + str + " ");
        } else {
          str = "?n" + id++;
          out.print("(let (" + str + " ");
        } 

        assert (table.get(cur.getId()) == null);
        table.put (cur.getId(), str);

        out.print ("(");
        assert (cur.toString() != null);
        out.print (cur.toString());
        children = cur.getChildren();
        assert (children != null);
        for (SMTNode child : children){ 

          if ((temp = subst.get(child.getId())) != null)
            child = temp;

          str = table.get(child.getId());
          assert (str != null);
          out.print(" ");
          out.print(str);
        }
        out.println("))");
      }
    }
    out.println(table.get(root.getId()));
    for (int i = 0; i < id; i++)
      out.print(")");
    out.println("");
    SMTNodeManager.setMark((byte) 0, this.formula);
    /* clear mark flags of substitutions */
    for (SMTNode n : subst.values())
      SMTNodeManager.setMark((byte) 0, n);
  }

  public void dump (PrintStream out){
    assert (out != null);
    assert (this.formula != null);
    assert (this.formula.getType() == SMTTypeManager.booleanType);
    this.dumpSubstitute (out, new HashMap<Long, SMTNode>());
  }


  /* result starts with root formula and performs a BFS
   * note that formulas may also occur inside ite */
  public SMTNode [] getFormulas () {
    Queue<SMTNode> queue, resultQueue;
    SMTNode cur;
    SMTNode []children;
    SMTNode []result;

    assert (this.formula != null);

    resultQueue = new LinkedList<SMTNode> ();
    queue = new LinkedList<SMTNode> ();
    queue.add (this.formula);
    while ((cur = queue.poll ()) != null) {

      assert (cur.getMark() == (byte) 0 || cur.getMark() == (byte) 1);
      if (cur.getMark() == (byte) 1)
        continue;

      cur.setMark ((byte) 1);

      if (cur.getType () == SMTTypeManager.booleanType)
        resultQueue.add (cur);

      children = cur.getChildren();
      if (children != null) {
        for (SMTNode child : children)
          queue.add (child);
      }
      
    }

    /* cleanup mark flags */
    SMTNodeManager.setMark ((byte) 0, this.formula);

    result = new SMTNode[resultQueue.size()];
    return resultQueue.toArray(result);
  }

}
