/*  DeltaSMT: Delta-debugger for Satisfiablity Modulo Theories (SMT) benchmarks.
 *  Copyright (C) 2009  Robert Daniel Brummayer
 *
 *  This file is part of DeltaSMT.
 *
 *  DeltaSMT is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DeltaSMT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.*;
import java.math.*;

public class SMTNodeManager
{

  protected SMTTypeManager typeMgr;

  protected HashMap<String, SMTNode> nodes;

  protected Stack<SMTNode> recordedNodes;

  protected boolean recordingEnabled;

  /* normalize before hashing, e.g. use commutativity:
   * (bvand e0 e1) and (bvand e1 e0) are mapped to the same node */
  protected boolean normSharing;

  public static final SMTNode TRUE = new SMTNode (SMTNodeKind.TRUE, 
                                                  SMTTypeManager.booleanType);

  public static final SMTNode FALSE = new SMTNode (SMTNodeKind.FALSE, 
                                                   SMTTypeManager.booleanType);

  protected void newNode (String hashKey, SMTNode node) {
    assert (hashKey != null);
    assert (node != null);
    this.nodes.put(hashKey, node);
    if (this.recordingEnabled)
      {
        assert (this.recordedNodes != null);
        this.recordedNodes.push (node);
      }
  }

  public SMTNodeManager (SMTTypeManager typeMgr){
    assert (typeMgr != null);

    this.typeMgr = typeMgr;
    this.nodes = new HashMap<String, SMTNode>();
    this.recordedNodes = null;
    this.recordingEnabled = false;
    this.normSharing = true;
  }

  public SMTNodeManager (SMTTypeManager typeMgr, boolean normSharing){
    assert (typeMgr != null);

    this.typeMgr = typeMgr;
    this.nodes = new HashMap<String, SMTNode>();
    this.recordedNodes = null;
    this.recordingEnabled = false;
    this.normSharing = normSharing;
  }

  public ConstBV bvconst (int width, BigInteger value) {
    SMTNode result;
    BVType type;
    String hashKey;

    assert (width > 0);
    assert (value.compareTo(BigInteger.ZERO) > -1);

    type = typeMgr.BVType(width);
    hashKey = type.toString() + " " + value.toString();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new ConstBV (type, value);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return (ConstBV) result;
  }

  public Extrafun extrafun (SMTType type, String name, SMTNode []children) {
    String hashKey;
    SMTNode result;
    StringBuilder builder;

    assert (type != null);
    assert (name != null);

    builder = new StringBuilder();
    builder.append (name);
    if (children != null) {
      for (SMTNode child : children) {
        builder.append (" ");
        builder.append (child.getId());
      }
    }
    hashKey = builder.toString();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new Extrafun (type, name, children);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return (Extrafun) result;
  }

  public Extrapred extrapred (String name, SMTNode []children) {
    String hashKey;
    SMTNode result;
    StringBuilder builder;

    assert (name != null);

    builder = new StringBuilder();
    builder.append (name);
    if (children != null) {
      for (SMTNode child : children) {
        builder.append (" ");
        builder.append (child.getId());
      }
    }
    hashKey = builder.toString();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new Extrapred (name, children);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return (Extrapred) result;
  }

  protected SMTNode unarySMTNode (SMTNodeKind kind, SMTType type, SMTNode n){
    assert (kind != null);
    assert (type != null);
    assert (n != null);
    assert (kind.getArity() == 1);
    assert (!SMTNodeKind.isIndexedUnaryKind(kind));

    String hashKey = kind.toString() + " " + n.getId(); 
    SMTNode result = this.nodes.get(hashKey);
    if (result == null) {
      result = new SMTNode (kind, type, n);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return result;
  }

  protected SMTNode indexedUnaryBVSMTNode (SMTNodeKind kind, BVType type, 
                                           SMTNode n, int value){
    assert (kind != null);
    assert (type != null);
    assert (n != null);
    assert (SMTNodeKind.isIndexedUnaryKind(kind));
    assert (kind != SMTNodeKind.EXTRACT);
    assert (value >= 0);

    String hashKey = kind.toString() + " " + n.getId() + " " + value; 
    SMTNode result = this.nodes.get(hashKey);
    if (result == null) {
      result = new IndexedUnaryBVNode (kind, type, n, value);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return result;
  }

  protected SMTNode indexedUnaryBVSMTNodeExtract (BVType type, SMTNode n, 
                                                  int upper, int lower){
    assert (type != null);
    assert (n != null);
    assert (lower >= 0);
    assert (upper >= lower);

    String hashKey = SMTNodeKind.EXTRACT.toString() + " " + n.getId() + " " + 
                     upper + " " + lower;
    SMTNode result = this.nodes.get(hashKey);
    if (result == null) {
      result = new Extract (type, n, upper, lower);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return result;
  }



  public SMTNode not (SMTNode n0) throws TypeException {
    SMTNode result;

    assert (n0 != null);
    assert (n0.getType() != null);

    if (n0.getType() != SMTTypeManager.booleanType)
      throw new TypeException ("incompatible types: expected: " + 
                                SMTTypeManager.booleanType.toString() +
                                " but was: " + n0.getType().toString());

    result = unarySMTNode (SMTNodeKind.NOT, n0.getType(), n0);
    return result;
  }

  public SMTNode bvnot (SMTNode n0) throws TypeException {
    SMTNode result;

    assert (n0 != null);
    assert (n0.getType() != null);

    if (!(n0.getType() instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" + 
                               " but was: " + n0.getType().toString());

    result = unarySMTNode (SMTNodeKind.BVNOT, n0.getType(), n0);
    return result;
  }

  public SMTNode bvneg (SMTNode n0) throws TypeException {
    SMTNode result;

    assert (n0 != null);
    assert (n0.getType() != null);

    if (!(n0.getType() instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" + 
                               " but was: " + n0.getType().toString());

    result = unarySMTNode (SMTNodeKind.BVNEG, n0.getType(), n0);
    return result;
  }

  public SMTNode extract (SMTNode n0, int upper, int lower) 
    throws TypeException {
    SMTNode result;
    BVType type;

    assert (n0 != null);
    assert (n0.getType() != null);
    assert (lower >= 0);
    assert (upper >= lower);

    if (!(n0.getType() instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" + 
                               " but was: " + n0.getType().toString());

    assert (upper < ((BVType)n0.getType()).getWidth());
    type = this.typeMgr.BVType(upper - lower + 1);
    result = indexedUnaryBVSMTNodeExtract (type, n0, upper, lower);
    return result;
  }


  public SMTNode repeat (SMTNode n0, int value) throws TypeException {
    SMTNode result;
    BVType type;

    assert (n0 != null);
    assert (n0.getType() != null);
    assert (value > 0);

    if (!(n0.getType() instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" + 
                               " but was: " + n0.getType().toString());

    type = this.typeMgr.BVType(((BVType)n0.getType()).getWidth() * value);
    result = indexedUnaryBVSMTNode (SMTNodeKind.REPEAT, type, n0, value);
    return result;
  }

  public SMTNode zero_extend (SMTNode n0, int value) throws TypeException {
    SMTNode result;
    BVType type;

    assert (n0 != null);
    assert (n0.getType() != null);
    assert (value >= 0);

    if (!(n0.getType() instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" + 
                               " but was: " + n0.getType().toString());

    type = this.typeMgr.BVType(((BVType)n0.getType()).getWidth() + value);
    result = indexedUnaryBVSMTNode (SMTNodeKind.ZERO_EXTEND, type, n0, value);
    return result;
  }

  public SMTNode sign_extend (SMTNode n0, int value) throws TypeException {
    SMTNode result;
    BVType type;

    assert (n0 != null);
    assert (n0.getType() != null);
    assert (value >= 0);

    if (!(n0.getType() instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" + 
                               " but was: " + n0.getType().toString());

    type = this.typeMgr.BVType(((BVType)n0.getType()).getWidth() + value);
    result = indexedUnaryBVSMTNode (SMTNodeKind.SIGN_EXTEND, type, n0, value);
    return result;
  }

  public SMTNode rotate_left (SMTNode n0, int value) throws TypeException {
    SMTNode result;

    assert (n0 != null);
    assert (n0.getType() != null);
    assert (value >= 0);

    if (!(n0.getType() instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" + 
                               " but was: " + n0.getType().toString());

    assert (value < ((BVType)n0.getType()).getWidth());
    result = indexedUnaryBVSMTNode (SMTNodeKind.ROTATE_LEFT, 
                                    (BVType) n0.getType(), n0, value);
    return result;
  }

  public SMTNode rotate_right (SMTNode n0, int value) throws TypeException {
    SMTNode result;

    assert (n0 != null);
    assert (n0.getType() != null);
    assert (value >= 0);

    if (!(n0.getType() instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" + 
                               " but was: " + n0.getType().toString());

    assert (value < ((BVType)n0.getType()).getWidth());
    result = indexedUnaryBVSMTNode (SMTNodeKind.ROTATE_RIGHT, 
                                    (BVType) n0.getType(), n0, value);
    return result;
  }


  protected SMTNode binarySMTNode (SMTNodeKind kind, SMTType type, SMTNode n0, 
                                   SMTNode n1){
    SMTNode temp, result;
    String hashKey;
    assert (kind != null);
    assert (kind.getArity() == 2);
    assert (type != null);
    assert (n0 != null);
    assert (n1 != null);

    if (this.normSharing && kind.isCommutative()){
      /* normalize commutative operators */
      if (n1.compareTo(n0) == -1){
        temp = n0;
        n0 = n1;
        n1 = temp;
      } 
    }

    hashKey = kind.toString() + " " + n0.getId() + " " + n1.getId();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new SMTNode (kind, type, n0, n1);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return result;
  }

  protected SMTNode binaryBooleanSMTNode (SMTNodeKind kind, SMTNode n0, 
                                          SMTNode n1)
  throws TypeException {
    assert (kind != null);
    assert (n0 != null);
    assert (n1 != null);
    assert (n0.getType() != null);
    assert (n1.getType() != null);

    if (n0.getType() != SMTTypeManager.booleanType)
      throw new TypeException ("incompatible types: expected: " + 
                                SMTTypeManager.booleanType.toString() +
                                " but first operand was: " + 
                                  n0.getType().toString());

    if (n1.getType() != SMTTypeManager.booleanType)
      throw new TypeException ("incompatible types: expected: " + 
                                SMTTypeManager.booleanType.toString() +
                                " but second operand was: " + 
                                n1.getType().toString());

    return binarySMTNode (kind, SMTTypeManager.booleanType, n0, n1);
  }

  public SMTNode nAryBooleanSMTNode (SMTNodeKind kind, SMTNode []nodes) 
    throws TypeException {
    SMTType curType;
    SMTNode cur, result;
    SMTNode []temp;
    StringBuilder builder;
    String hashKey;

    assert (kind != null);
    assert (kind == SMTNodeKind.AND || kind == SMTNodeKind.OR);
    assert (nodes != null);
    assert (nodes.length >= 2);
    assert (nodes[0] != null);
    assert (nodes[1] != null);

    /* build temporary array and do type checking */
    temp = new SMTNode[nodes.length];
    temp[0] = nodes[0];
    for (int i = 1; i < nodes.length; i++) {
      cur = nodes[i];
      assert (cur != null);
      curType = cur.getType();
      assert (curType != null);
      if (curType != SMTTypeManager.booleanType)
        throw new TypeException ("incompatible types: expected " + 
                                 SMTTypeManager.booleanType.toString() + 
                                 " but " + i + ". operand was: " + 
                                 curType.toString());
      temp[i] = nodes[i];
    }

    /* normalize ? */
    if (this.normSharing)
      Arrays.sort(temp);
    
    /* build hash key */
    builder = new StringBuilder();
    builder.append(kind.toString());
    for (int i = 0; i < temp.length; i++) {
      cur = temp[i];
      assert (cur != null);
      builder.append(" ");
      builder.append(cur.getId());
    }

    hashKey = builder.toString();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new SMTNode (kind, SMTTypeManager.booleanType, temp);
      this.newNode (hashKey, result);
    } else { 
      result.incRefCounter();
    }

    return result;
  }


  public SMTNode and (SMTNode []nodes) throws TypeException {
    assert (nodes != null);
    return nAryBooleanSMTNode (SMTNodeKind.AND, nodes);
  }

  public SMTNode or (SMTNode []nodes) throws TypeException {
    assert (nodes != null);
    return nAryBooleanSMTNode (SMTNodeKind.OR, nodes);
  }

  public SMTNode implies (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBooleanSMTNode (SMTNodeKind.IMPLIES, n0, n1);
  }

  public SMTNode xor (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBooleanSMTNode (SMTNodeKind.XOR, n0, n1);
  }

  public SMTNode iff (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBooleanSMTNode (SMTNodeKind.IFF, n0, n1);
  }

  protected SMTNode binaryBVSMTNode (SMTNodeKind kind, SMTNode n0, SMTNode n1)
  throws TypeException {
    SMTType typeN0, typeN1;

    assert (kind != null);
    assert (n0 != null);
    assert (n1 != null);

    typeN0 = n0.getType();
    typeN1 = n1.getType();

    assert (typeN0 != null);
    assert (typeN1 != null);

    if (!(typeN0 instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" +
                                " but first operand was: " + typeN0.toString());

    if (!(typeN1 instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" +
                                " but second operand was: " + 
                                typeN1.toString());

    if (typeN0 != typeN1 && kind != SMTNodeKind.CONCAT)
      throw new TypeException ("incompatible types: " + typeN0.toString() +
                                " and " + typeN1.toString());

    return binarySMTNode (kind, typeN0, n0, n1);
  }

  public SMTNode bvadd (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVADD, n0, n1);
  }

  public SMTNode bvmul (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVMUL, n0, n1);
  }

  public SMTNode bvand (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVAND, n0, n1);
  }

  public SMTNode bvnand (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVNAND, n0, n1);
  }

  public SMTNode bvor (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVOR, n0, n1);
  }

  public SMTNode bvnor (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVNOR, n0, n1);
  }

  public SMTNode bvxor (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVXOR, n0, n1);
  }

  public SMTNode bvxnor (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVXNOR, n0, n1);
  }

  public SMTNode bvcomp (SMTNode n0, SMTNode n1) throws TypeException {
    SMTNode result;
    assert (n0 != null);
    assert (n1 != null);
    result = binaryBVSMTNode (SMTNodeKind.BVCOMP, n0, n1);
    /* adjust type */
    result.setType (this.typeMgr.BVType(1));
    return result;
  }

  public SMTNode bvshl (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVSHL, n0, n1);
  }

  public SMTNode bvlshr (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVLSHR, n0, n1);
  }

  public SMTNode bvashr (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVASHR, n0, n1);
  }

  public SMTNode bvsub (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVSUB, n0, n1);
  }

  public SMTNode bvudiv (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVUDIV, n0, n1);
  }

  public SMTNode bvurem (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVUREM, n0, n1);
  }

  public SMTNode bvsdiv (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVSDIV, n0, n1);
  }

  public SMTNode bvsrem (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVSREM, n0, n1);
  }

  public SMTNode bvsmod (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVSMTNode (SMTNodeKind.BVSMOD, n0, n1);
  }

  public SMTNode concat (SMTNode n0, SMTNode n1) throws TypeException {
    SMTNode result;
    BVType typeN0, typeN1;

    assert (n0 != null);
    assert (n1 != null);

    result =  binaryBVSMTNode (SMTNodeKind.CONCAT, n0, n1);
    /* adjust type */
    typeN0 = (BVType) n0.getType();
    typeN1 = (BVType) n1.getType();
    result.setType (this.typeMgr.BVType(typeN0.getWidth() + typeN1.getWidth()));
    return result;
  }

  public SMTNode eq (SMTNode n0, SMTNode n1) throws TypeException {
    SMTType typeN0, typeN1;
    assert (n0 != null);
    assert (n1 != null);

    typeN0 = n0.getType();
    typeN1 = n1.getType();

    assert (typeN0 != null);
    assert (typeN1 != null);

    if (typeN0 != typeN1)
      throw new TypeException ("incompatible types: " + typeN0.toString() +
                                " and " + typeN1.toString());

    return binarySMTNode (SMTNodeKind.EQ, SMTTypeManager.booleanType, n0, n1);
  }

  protected SMTNode binaryBVPredSMTNode (SMTNodeKind kind, SMTNode n0, 
                                         SMTNode n1)
  throws TypeException {
    SMTType typeN0, typeN1;

    assert (kind != null);
    assert (n0 != null);
    assert (n1 != null);

    typeN0 = n0.getType();
    typeN1 = n1.getType();

    assert (typeN0 != null);
    assert (typeN1 != null);

    if (!(typeN0 instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" +
                                " but first operand was: " + typeN0.toString());

    if (!(typeN1 instanceof BVType))
      throw new TypeException ("incompatible types: expected: (BV <n>)" +
                                " but second operand was: " + 
                                typeN1.toString());

    if (typeN0 != typeN1)
      throw new TypeException ("incompatible types: " + typeN0.toString() +
                                " and " + typeN1.toString());

      
    return binarySMTNode (kind, SMTTypeManager.booleanType, n0, n1);
  }

  public SMTNode bvult (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVPredSMTNode (SMTNodeKind.BVULT, n0, n1);
  }

  public SMTNode bvule (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVPredSMTNode (SMTNodeKind.BVULE, n0, n1);
  }

  public SMTNode bvugt (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVPredSMTNode (SMTNodeKind.BVUGT, n0, n1);
  }

  public SMTNode bvuge (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVPredSMTNode (SMTNodeKind.BVUGE, n0, n1);
  }

  public SMTNode bvslt (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVPredSMTNode (SMTNodeKind.BVSLT, n0, n1);
  }

  public SMTNode bvsle (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVPredSMTNode (SMTNodeKind.BVSLE, n0, n1);
  }

  public SMTNode bvsgt (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVPredSMTNode (SMTNodeKind.BVSGT, n0, n1);
  }

  public SMTNode bvsge (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryBVPredSMTNode (SMTNodeKind.BVSGE, n0, n1);
  }

  public SMTNode select (SMTNode n0, SMTNode n1) throws TypeException {
    SMTType typeN0, typeN1, indexType, elementType;

    assert (n0 != null);
    assert (n1 != null);

    typeN0 = n0.getType();
    typeN1 = n1.getType();
    if (!(typeN0 instanceof ArrayType))
      throw new TypeException ("incompatible types: expected: (Array)" +
                                " but first operand was: " + 
                                typeN0.toString());
    indexType = ((ArrayType) typeN0).getIndexType();
    elementType = ((ArrayType) typeN0).getElementType();
    if (indexType != typeN1)
      throw new TypeException ("incomptaible types: type of select index " +
                                "does not match index type of array term");

    return binarySMTNode (SMTNodeKind.SELECT, elementType, n0, n1);
  }

  protected SMTNode ternarySMTNode (SMTNodeKind kind, SMTType type, SMTNode n0, 
                                    SMTNode n1, SMTNode n2){
    SMTNode result;
    String hashKey;
    assert (kind != null);
    assert (kind.getArity() == 3);
    assert (type != null);
    assert (n0 != null);
    assert (n1 != null);
    assert (n2 != null);

    hashKey = kind.toString() + " " + n0.getId() + " " + n1.getId() + " " + 
              n2.getId();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new SMTNode (kind, type, n0, n1, n2);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return result;
  }

  public SMTNode if_then_else (SMTNode n0, SMTNode n1, SMTNode n2) 
    throws TypeException {

    assert (n0 != null);
    assert (n1 != null);
    assert (n2 != null);
    assert (n0.getType() != null);
    assert (n1.getType() != null);
    assert (n2.getType() != null);

    if (n0.getType() != SMTTypeManager.booleanType)
      throw new TypeException ("incompatible types: expected: " + 
                                SMTTypeManager.booleanType.toString() +
                                " but first operand was: " + 
                                n0.getType().toString());

    if (n1.getType() != SMTTypeManager.booleanType)
      throw new TypeException ("incompatible types: expected: " + 
                                SMTTypeManager.booleanType.toString() +
                                " but second operand was: " + 
                                n1.getType().toString());

    if (n2.getType() != SMTTypeManager.booleanType)
      throw new TypeException ("incompatible types: expected: " + 
                                SMTTypeManager.booleanType.toString() +
                                " but third operand was: " + 
                                n2.getType().toString());

    return ternarySMTNode (SMTNodeKind.IF_THEN_ELSE, SMTTypeManager.booleanType,
                           n0, n1, n2);
  }

  public SMTNode ite (SMTNode n0, SMTNode n1, SMTNode n2) 
    throws TypeException {
    SMTType typeN0, typeN1, typeN2;

    assert (n0 != null);
    assert (n1 != null);
    assert (n2 != null);

    typeN0 = n0.getType();
    typeN1 = n1.getType();
    typeN2 = n2.getType();

    assert (typeN0 != null);
    assert (typeN1 != null);
    assert (typeN2 != null);

    if (typeN0 != SMTTypeManager.booleanType)
      throw new TypeException ("incompatible types: expected: " + 
                                SMTTypeManager.booleanType.toString() +
                                " but first operand was: " + typeN0.toString());
    if (typeN1 != typeN2)
      throw new TypeException ("incompatible types: second operand: " + 
                                typeN1.toString() + " third operand: " + 
                                typeN2.toString());
    return ternarySMTNode (SMTNodeKind.ITE, typeN1, n0, n1, n2);
  }

  public SMTNode store (SMTNode n0, SMTNode n1, SMTNode n2) 
    throws TypeException {
    SMTType typeN0, typeN1, typeN2, indexType, elementType;

    assert (n0 != null);
    assert (n1 != null);
    assert (n2 != null);

    typeN0 = n0.getType();
    typeN1 = n1.getType();
    typeN2 = n2.getType();

    assert (typeN0 != null);
    assert (typeN1 != null);
    assert (typeN2 != null);

    if (!(typeN0 instanceof ArrayType))
      throw new TypeException ("incompatible types: expected: (Array)" +
                                " but first operand was: " + 
                                typeN0.toString());

    indexType = ((ArrayType) typeN0).getIndexType();
    elementType = ((ArrayType) typeN0).getElementType();

    if (indexType != typeN1)
      throw new TypeException ("incomptaible types: type of store index " +
                                "does not match index type of array term");
    if (elementType != typeN2)
      throw new TypeException ("incomptaible types: type of store element " +
                                "does not match element type of array term");

    return ternarySMTNode (SMTNodeKind.STORE, typeN0, n0, n1, n2);
  }


  public SMTNode distinct (SMTNode[] nodes) throws TypeException {
    SMTType typeN0, curType;
    SMTNode cur, result;
    SMTNode []temp;
    StringBuilder builder;
    String hashKey;

    assert (nodes != null);
    assert (nodes.length >= 2);
    assert (nodes[0] != null);

    typeN0 = nodes[0].getType();
    assert (typeN0 != null);

    /* build temporary array and do type checking */
    temp = new SMTNode[nodes.length];
    temp[0] = nodes[0];
    for (int i = 1; i < nodes.length; i++) {
      cur = nodes[i];
      assert (cur != null);
      curType = cur.getType();
      assert (curType != null);
      if (typeN0 != curType)
        throw new TypeException ("incompatible types: 1. operand: " + 
                                 typeN0.toString() + " " + (i+1) + 
                                 ". operand: " + curType.toString());
      temp[i] = nodes[i];
    }

    /* normalize ? */
    if (this.normSharing)
      Arrays.sort(temp);
    
    /* build hash key */
    builder = new StringBuilder();
    builder.append(SMTNodeKind.DISTINCT.toString());
    for (int i = 0; i < temp.length; i++) {
      cur = temp[i];
      assert (cur != null);
      builder.append(" ");
      builder.append(cur.getId());
    }

    hashKey = builder.toString();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new SMTNode (SMTNodeKind.DISTINCT, SMTTypeManager.booleanType, 
                            temp);
      this.newNode (hashKey, result);
    } else { 
      result.incRefCounter();
    }

    return result;
  }

  @SuppressWarnings("unchecked")
  public ConstTerm<BigInteger> intconst (BigInteger value) {
    SMTNode result;
    String hashKey;

    assert (value.compareTo(BigInteger.ZERO) > -1);

    hashKey = SMTTypeManager.integerType.toString() + " " + value.toString();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new ConstTerm<BigInteger> (SMTTypeManager.integerType, value);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return (ConstTerm<BigInteger>) result;
  }

  @SuppressWarnings("unchecked")
  public ConstTerm<BigInteger> intconstasreal (BigInteger value) {
    SMTNode result;
    String hashKey;

    assert (value.compareTo(BigInteger.ZERO) > -1);

    hashKey = SMTTypeManager.realType.toString() + " " + value.toString();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new ConstTerm<BigInteger> (SMTTypeManager.realType, value);
      this.newNode (hashKey, result);
    } else {
      result.incRefCounter();
    }

    return (ConstTerm<BigInteger>) result;
  }



  public SMTNode binaryInterpretedPredSMTNode (SMTNodeKind kind, SMTNode n0, 
                                               SMTNode n1) throws TypeException{
    SMTType typeN0, typeN1;

    assert (kind != null);
    assert (n0 != null);
    assert (n1 != null);
    assert (kind == SMTNodeKind.LT || kind == SMTNodeKind.GT || 
            kind == SMTNodeKind.LE || kind == SMTNodeKind.GE);

    typeN0 = n0.getType();
    typeN1 = n1.getType();

    if (typeN0 != SMTTypeManager.integerType &&
        typeN0 != SMTTypeManager.realType)
      throw new TypeException ("incompatible types: expected: (Int) or (Real)" +
                                " but first operand was: " + typeN0.toString());
    if (typeN1 != SMTTypeManager.integerType &&
        typeN1 != SMTTypeManager.realType)
      throw new TypeException ("incompatible types: expected: (Int) or (Real)" +
                                " but second operand was: " + 
                                typeN1.toString());

    return binarySMTNode (kind, SMTTypeManager.booleanType, n0, n1);
  }

  public SMTNode lt (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryInterpretedPredSMTNode (SMTNodeKind.LT, n0, n1);
  }

  public SMTNode gt (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryInterpretedPredSMTNode (SMTNodeKind.GT, n0, n1);
  }

  public SMTNode le (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryInterpretedPredSMTNode (SMTNodeKind.LE, n0, n1);
  }

  public SMTNode ge (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryInterpretedPredSMTNode (SMTNodeKind.GE, n0, n1);
  }

  public SMTNode neg (SMTNode n0) throws TypeException {
    SMTNode result;

    assert (n0 != null);
    assert (n0.getType() != null);

    if (n0.getType() != SMTTypeManager.integerType && 
        n0.getType() != SMTTypeManager.realType)
      throw new TypeException ("incompatible types: expected: (Int) or (Real)"+ 
                               " but was: " + n0.getType().toString());

    result = unarySMTNode (SMTNodeKind.NEG, n0.getType(), n0);
    return result;
  }

  public SMTNode binaryInterpretedFuncSMTNode (SMTNodeKind kind, SMTNode n0, 
                                               SMTNode n1) throws TypeException{
    SMTType typeN0, typeN1;

    assert (kind != null);
    assert (n0 != null);
    assert (n1 != null);
    assert (kind == SMTNodeKind.ADD || kind == SMTNodeKind.SUB || 
            kind == SMTNodeKind.MUL || kind == SMTNodeKind.DIV);

    typeN0 = n0.getType();
    typeN1 = n1.getType();

    if (typeN0 != SMTTypeManager.integerType &&
        typeN0 != SMTTypeManager.realType)
      throw new TypeException ("incompatible types: expected: (Int) or (Real)" +
                                " but first operand was: " + typeN0.toString());
    if (typeN1 != SMTTypeManager.integerType &&
        typeN1 != SMTTypeManager.realType)
      throw new TypeException ("incompatible types: expected: (Int) or (Real)" +
                                " but second operand was: " + 
                                typeN1.toString());
    if (typeN0 != typeN1)
      throw new TypeException ("incompatible types: " + typeN0 + " " + typeN1);

    return binarySMTNode (kind, typeN0, n0, n1);
  }


  public SMTNode sub (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryInterpretedFuncSMTNode (SMTNodeKind.SUB, n0, n1);
  }

  public SMTNode mul (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryInterpretedFuncSMTNode (SMTNodeKind.MUL, n0, n1);
  }

  public SMTNode div (SMTNode n0, SMTNode n1) throws TypeException {
    assert (n0 != null);
    assert (n1 != null);
    return binaryInterpretedFuncSMTNode (SMTNodeKind.DIV, n0, n1);
  }

  public SMTNode add (SMTNode []nodes) throws TypeException {
    SMTType typeN0, curType;
    SMTNode cur, result;
    SMTNode []temp;
    StringBuilder builder;
    String hashKey;

    assert (nodes != null);
    assert (nodes.length >= 1); /* (+ x) special case in QF_RDL */
    assert (nodes[0] != null);

    typeN0 = nodes[0].getType();
    assert (typeN0 != null);

    /* build temporary array and do type checking */
    temp = new SMTNode[nodes.length];
    temp[0] = nodes[0];
    for (int i = 1; i < nodes.length; i++) {
      cur = nodes[i];
      assert (cur != null);
      curType = cur.getType();
      assert (curType != null);
      if (typeN0 != curType)
        throw new TypeException ("incompatible types: 1. operand: " + 
                                 typeN0.toString() + " " + (i+1) + 
                                 ". operand: " + curType.toString());
      temp[i] = nodes[i];
    }

    /* normalize ? */
    if (this.normSharing)
      Arrays.sort(temp);
    
    /* build hash key */
    builder = new StringBuilder();
    builder.append(SMTNodeKind.ADD.toString());
    for (int i = 0; i < temp.length; i++) {
      cur = temp[i];
      assert (cur != null);
      builder.append(" ");
      builder.append(cur.getId());
    }

    hashKey = builder.toString();
    result = this.nodes.get(hashKey);
    if (result == null) {
      result = new SMTNode (SMTNodeKind.ADD, typeN0, temp);
      this.newNode (hashKey, result);
    } else { 
      result.incRefCounter();
    }

    return result;
  }


  public SMTNode copy (SMTNode n){
    assert (n != null);
    n.incRefCounter();
    return n;
  }

  public void release (SMTNode n){
    Stack<SMTNode> stack;
    SMTNode cur;
    SMTNode []children;

    assert (n != null);
    assert (n.refs > 0);

    if (n != TRUE && n != FALSE) {
      assert (this.nodes.get (n.hashKey()) != null);
      if (n.refs > 1) {
        n.refs--;
      } else {
        /* recursively release children */
        stack = new Stack<SMTNode>();
        stack.push (n);
        do {
          cur = stack.pop();
          if (cur.refs > 1) {
            cur.refs--;
          } else {
            assert (cur.refs == 1);
            children = cur.getChildren();
            if (children != null) {
              for (int i = 0; i < children.length; i++)
                stack.push(children[i]);
            }
            /* remove from hashtable */
            cur.refs = 0;
            if (this.recordingEnabled)
              this.recordedNodes.remove (cur.hashKey());
            this.nodes.remove (cur.hashKey());
          }
        } while (!stack.empty());
      }
    }
  }

  public static void setMark(byte mark, SMTNode node){
    Stack<SMTNode> stack = new Stack<SMTNode>();
    SMTNode []children;
    SMTNode cur;

    assert (node != null);

    stack.push(node);
    do {
      cur = stack.pop();
      if (cur.mark != mark){
        cur.mark = mark;
        children = cur.getChildren();
        if (children != null){
          for (int i = 0; i < children.length; i++)
            stack.push (children[i]);
        } 
      }
    } while (!stack.empty());
  }

  public int numberOfNodes (){
    return this.nodes.size();
  }

  public SMTNode zero (SMTType t){
    assert (t != null);

    if (t == SMTTypeManager.booleanType)
      return SMTNodeManager.FALSE;

    if (t instanceof BVType)
      return this.bvconst (((BVType) t).getWidth(), BigInteger.ZERO);

    if (t == SMTTypeManager.integerType)
      return intconst (BigInteger.ZERO);

    /* support other types here */

    return null;
  }

  public SMTNode one (SMTType t){
    assert (t != null);

    if (t == SMTTypeManager.booleanType)
      return SMTNodeManager.TRUE;

    if (t instanceof BVType)
      return this.bvconst (((BVType) t).getWidth(), BigInteger.ONE);

    if (t == SMTTypeManager.integerType)
      return intconst (BigInteger.ONE);

    /* support other types here */

    return null;
  }

  public void startNodeRecording (){
    this.recordingEnabled = true;
    this.recordedNodes = new Stack<SMTNode>();
  }

  public Stack<SMTNode> collectRecordedNodesAndStopRecording () {
    Stack<SMTNode> result = this.recordedNodes;
    this.recordedNodes = null;
    this.recordingEnabled = false;
    return result;
  }

  public SMTTypeManager getTypeManager() {
    return this.typeMgr;
  }


}
